
from .base import Command


class QPIGS2(Command):
    REQUEST_FMT = 'QPIGS2'


class QPIGS2_MAX(QPIGS2):
    RESPONSE_FMT = (
        '{pv2_current:f} '
        '{pv2_voltage:f} '
        '{pv2_power:f}'
    )

    UNITS = {
        'pv2_current': 'A',
        'pv2_voltage': 'V',
        'pv2_power': 'W'
    }

    DISPLAY_NAMES = {
        'pv2_current': 'PV2 Current',
        'pv2_voltage': 'PV2 Voltage',
        'pv2_power': 'PV2 Power'
    }

    VIRTUAL_PROPERTIES = {
        'pv2_power': {
            'value': lambda properties: properties.get('pv2_current', 0) * properties.get('pv2_voltage', 0),
            'type': 'float'
        }
    }




from .base import Command


class QPIGS(Command):
    REQUEST_FMT = 'QPIGS'

    RESPONSE_FMT = (
        '{grid_voltage:f} '
        '{grid_frequency:f} '
        '{ac_output_voltage:f} '
        '{ac_output_frequency:f} '
        '{ac_output_apparent_power:f} '
        '{ac_output_active_power:f} '
        '{ac_output_load:f} '
        '{bus_voltage:f} '
        '{battery_voltage:f} '
        '{battery_charging_current:f} '
        '{battery_state_of_charge:f} '
        '{heat_sink_temperature:f} '
        '{pv_current:f} '
        '{pv_voltage:f} '
        '{scc_voltage:f} '
        '{battery_discharging_current:f} '
        '{has_sbu_priority:b}'
        '{is_configuration_status_changed:b}'
        '{is_scc_firmware_updated:b}'
        '{has_load:b}'
        '{is_battery_voltage_too_steady_while_charging:b}'
        '{is_battery_charging:b}'
        '{is_battery_charging_from_scc:b}'
        '{is_battery_charging_from_grid:b}'
    )

    UNITS = {
        'ac_output_active_power': 'W',
        'ac_output_apparent_power': 'VA',
        'ac_output_frequency': 'Hz',
        'ac_output_load': '%',
        'ac_output_voltage': 'V',
        'battery_charging_current': 'A',
        'battery_discharging_current': 'A',
        'battery_state_of_charge': '%',
        'battery_voltage': 'V',
        'battery_voltage_offset_fans': '10mV',
        'bus_voltage': 'V',
        'grid_frequency': 'Hz',
        'grid_voltage': 'V',
        'heat_sink_temperature': 'C',
        'pv_charging_power': 'W',
        'pv_current': 'A',
        'pv_power': 'W',
        'pv_voltage': 'V',
        'scc_voltage': 'V',
    }

    DISPLAY_NAMES = {
        'ac_output_active_power': 'AC Output Active Power',
        'ac_output_apparent_power': 'AC Output Apparent Power',
        'ac_output_frequency': 'AC Output Frequency',
        'ac_output_load': 'AC Output Load',
        'ac_output_voltage': 'AC Output Voltage',
        'battery_charging_current': 'Battery Charging Current',
        'battery_discharging_current': 'Battery Discharging Current',
        'battery_state_of_charge': 'Battery State Of Charge',
        'battery_voltage': 'Battery Voltage',
        'battery_voltage_offset_fans': 'Battery Voltage Offset Fans',
        'bus_voltage': 'Bus Voltage',
        'eeprom_version': 'EEPROM Version',
        'grid_frequency': 'Grid Frequency',
        'grid_voltage': 'Grid Voltage',
        'has_load': 'Has Load',
        'has_sbu_priority': 'Has SBU Priority',
        'heat_sink_temperature': 'Heat Sink Temperature',
        'is_ac_output_from_grid_or_pv': 'AC Output From Grid/PV',
        'is_battery_charging': 'Battery Charging',
        'is_battery_charging_from_grid': 'Battery Charging From Grid',
        'is_battery_charging_from_scc': 'Battery Charging From SCC',
        'is_battery_float_charging': 'Battery Float-charging',
        'is_battery_low': 'Battery Low',
        'is_battery_present': 'Battery Present',
        'is_battery_voltage_too_steady_while_charging': 'Battery Voltage Too Steady While Charging',
        'is_configuration_status_changed': 'Configuration Status Changed',
        'is_dustproof_installed': 'Dustproof Installed',
        'is_grid_present': 'Grid Present',
        'is_scc_active': 'SCC Active',
        'is_scc_firmware_updated': 'SCC Firmware Updated',
        'is_turned_on': 'Turned On',
        'pv_charging_power': 'PV Charging Power',
        'pv_current': 'PV Current',
        'pv_power': 'PV Power',
        'pv_voltage': 'PV Voltage',
        'scc_voltage': 'SCC Voltage',
    }

    VIRTUAL_PROPERTIES = {
        'pv_power': {
            'value': lambda properties: properties.get('pv_current', 0) * properties.get('pv_voltage', 0),
            'type': 'float'
        },
        'is_battery_charging': {
            'value': lambda properties: (
                properties.get('is_battery_charging_from_grid', False) or
                properties.get('is_battery_charging_from_scc', False)
            ),
            'type': 'bool'
        },
    }


class QPIGS_LV(QPIGS):
    RESPONSE_FMT = (
        '{grid_voltage:f} '
        '{grid_frequency:f} '
        '{ac_output_voltage:f} '
        '{ac_output_frequency:f} '
        '{ac_output_apparent_power:f} '
        '{ac_output_active_power:f} '
        '{ac_output_load:f} '
        '{bus_voltage:f} '
        '{battery_voltage:f} '
        '{battery_charging_current:f} '
        '{battery_state_of_charge:f} '
        '{heat_sink_temperature:f} '
        '{pv_current:f} '
        '{pv_voltage:f} '
        '{scc_voltage:f} '
        '{battery_discharging_current:f} '
        '{is_scc_active:b}'
        '{is_battery_charging_from_grid:b}'
        '{is_battery_charging_from_scc:b}'
        '{is_battery_low:b}'
        '{is_battery_present:b}'
        '{is_grid_present:b}'
        '{has_load:b} '
        '{_reserved_1:s} '
        '{_reserved_2:s} '
        '{pv_power:f} '
        '{is_battery_float_charging:b}'
        '{is_turned_on:b}'
        '{_reserved_3:b}'
    )


class QPIGS_GKMK(QPIGS):
    RESPONSE_FMT = (
        '{grid_voltage:f} '
        '{grid_frequency:f} '
        '{ac_output_voltage:f} '
        '{ac_output_frequency:f} '
        '{ac_output_apparent_power:f} '
        '{ac_output_active_power:f} '
        '{ac_output_load:f} '
        '{bus_voltage:f} '
        '{battery_voltage:f} '
        '{battery_charging_current:f} '
        '{battery_state_of_charge:f} '
        '{heat_sink_temperature:f} '
        '{pv_current:f} '
        '{pv_voltage:f} '
        '{scc_voltage:f} '
        '{battery_discharging_current:f} '
        '{is_ac_output_from_grid_or_pv:b}'
        '{is_configuration_status_changed:b}'
        '{is_scc_firmware_updated:b}'
        '{has_load:b}'
        '{_reserved_1:b}'
        '{is_battery_charging:b}'
        '{is_battery_charging_from_scc:b}'
        '{is_battery_charging_from_grid:b} '
        '{battery_voltage_offset_fans:d} '
        '{eeprom_version:d} '
        '{pv_charging_power:d} '
        '{is_battery_float_charging:b}'
        '{is_turned_on:b}'
        '{is_dustproof_installed:b}'
    )


class QPIGS_MAX(QPIGS):
    RESPONSE_FMT = (
        '{grid_voltage:f} '
        '{grid_frequency:f} '
        '{ac_output_voltage:f} '
        '{ac_output_frequency:f} '
        '{ac_output_apparent_power:f} '
        '{ac_output_active_power:f} '
        '{ac_output_load:f} '
        '{bus_voltage:f} '
        '{battery_voltage:f} '
        '{battery_charging_current:f} '
        '{battery_state_of_charge:f} '
        '{heat_sink_temperature:f} '
        '{pv_current:f} '
        '{pv_voltage:f} '
        '{scc_voltage:f} '
        '{battery_discharging_current:f} '
        '{has_sbu_priority:b}'
        '{is_configuration_status_changed:b}'
        '{is_scc_firmware_updated:b}'
        '{has_load:b}'
        '{is_battery_voltage_too_steady_while_charging:b}'
        '{is_battery_charging:b}'
        '{is_battery_charging_from_scc:b}'
        '{is_battery_charging_from_grid:b} '
        '{battery_voltage_offset_fans:d} '
        '{eeprom_version:d} '
        '{pv_charging_power:d} '
        '{is_battery_float_charging:b}'
        '{is_turned_on:b}'
        '{is_dustproof_installed:b}'
    )




from typing import List, Type

from .base import Command
from .mchgc import MCHGC
from .mnchgc import MNCHGC_Parallel, MNCHGC_Single
from .muchgc import MUCHGC_Single, MUCHGC_Parallel
from .pbatmaxdisc import PBATMAXDISC
from .pbcv import PBCV
from .pbdv import PBDV
from .pbft import PBFT
from .pbt import PBT
from .pcp import PCP
from .pcvv import PCVV
from .pop import POP
from .psdv import PSDV
from .qmchgcr import QMCHGCR
from .qmuchgcr import QMUCHGCR
from .qmod import QMOD
from .qpigs import QPIGS, QPIGS_LV, QPIGS_GKMK, QPIGS_MAX
from .qpigs2 import QPIGS2, QPIGS2_MAX
from .qpiri import QPIRI, QPIRI_GK, QPIRI_MAX, QPIRI_MK


COMMANDS_BY_MODEL = {
    'GK': [
        MNCHGC_Single,
        MUCHGC_Single,
        PBCV,
        PBDV,
        PBFT,
        PBT,
        PCP,
        PCVV,
        POP,
        PSDV,
        QMCHGCR,
        QMUCHGCR,
        QMOD,
        QPIGS_GKMK,
        QPIRI_GK,
    ],
    'LV': [
        MNCHGC_Single,
        PBCV,
        PBDV,
        PBFT,
        PBT,
        PCP,
        PCVV,
        POP,
        PSDV,
        QMCHGCR,
        QMOD,
        QPIGS_LV,
    ],
    'MAX': [
        MCHGC,
        MNCHGC_Parallel,
        MUCHGC_Parallel,
        PBATMAXDISC,
        PBCV,
        PBDV,
        PBFT,
        PBT,
        PCP,
        PCVV,
        POP,
        PSDV,
        QMCHGCR,
        QMOD,
        QMUCHGCR,
        QPIGS_MAX,
        QPIGS2_MAX,
        QPIRI_MAX,
    ],
    'MK': [
        MNCHGC_Parallel,
        MUCHGC_Parallel,
        PBCV,
        PBDV,
        PBFT,
        PBT,
        PCP,
        PCVV,
        POP,
        PSDV,
        QMCHGCR,
        QMOD,
        QPIGS_GKMK,
        QPIRI_MK,
    ],
}


def get_command_classes(model: str) -> List[Type[Command]]:
    return COMMANDS_BY_MODEL.get(model, [])




import abc
import asyncio
import logging

from typing import Dict, Optional

from qtoggleserver.lib.polled import PolledPeripheral

from .exceptions import MPPSolarTimeout
from .typing import Property


logger = logging.getLogger(__name__)


class MPPSolarInverter(PolledPeripheral, metaclass=abc.ABCMeta):
    DEFAULT_POLL_INTERVAL = 5
    RETRY_POLL_INTERVAL = 5

    TIMEOUT = 10

    logger = logger

    def __init__(
        self,
        *,
        model: str,
        **kwargs
    ) -> None:

        self._model: str = model
        self._properties: Dict[str, Property] = {}

        super().__init__(**kwargs)

    async def read_properties(self) -> None:
        raise NotImplementedError

    async def poll(self) -> None:
        try:
            await self.read_properties()

        except asyncio.TimeoutError as e:
            raise MPPSolarTimeout('Timeout reading inverter status') from e

    def get_property(self, name: str) -> Optional[Property]:
        return self._properties.get(name)

    async def set_property(self, name: str, value: Property) -> None:
        raise NotImplementedError




import abc

from typing import Any, cast, Dict, List, Optional

from qtoggleserver.core.typing import NullablePortValue, PortValue
from qtoggleserver.lib.polled import PolledPort

from .inverter import MPPSolarInverter


class MPPSolarPort(PolledPort, metaclass=abc.ABCMeta):
    def __init__(self, *, property_name: str, display_name: str, writable: bool, **kwargs) -> None:
        self._property_name: str = property_name

        super().__init__(**kwargs)

        self._display_name: str = display_name
        self._writable: bool = writable

    def get_peripheral(self) -> MPPSolarInverter:
        return cast(MPPSolarInverter, super().get_peripheral())

    def make_id(self) -> str:
        return self._property_name

    async def read_value(self) -> NullablePortValue:
        return self.get_peripheral().get_property(self._property_name)

    async def write_value(self, value: PortValue) -> None:
        await self.get_peripheral().set_property(self._property_name, value)


class BooleanPort(MPPSolarPort):
    TYPE = 'boolean'


class NumberPort(MPPSolarPort):
    TYPE = 'number'

    def __init__(self, *, unit: Optional[str] = None, choices: Optional[List[Dict[str, Any]]] = None, **kwargs) -> None:
        super().__init__(**kwargs)

        self._unit: Optional[str] = unit
        self._choices: Optional[List[Dict[str, Any]]] = choices


class StringPort(MPPSolarPort):
    TYPE = 'number'

    def __init__(self, *, unit: Optional[str] = None, choices: List[Dict[str, Any]], **kwargs) -> None:
        super().__init__(**kwargs)

        # Associate a number to each choice value, since we can't deal with string values
        self._value_mapping: Dict[str, int] = {}
        adapted_choices = []
        for i, choice in enumerate(choices):
            adapted_choice = dict(choice, value=i)
            adapted_choices.append(adapted_choice)
            self._value_mapping[choice['value']] = i

        self._unit: Optional[str] = unit
        self._choices: List[Dict[str, Any]] = adapted_choices

    async def read_value(self) -> NullablePortValue:
        value = self.get_peripheral().get_property(self._property_name)
        if value is not None:
            return self._value_mapping[value]




from .base import Command


class MUCHGC(Command):
    REQUEST_FMT = 'MUCHGC{battery_max_grid_charging_current:03.0f}'


class MUCHGC_Single(MUCHGC):
    REQUEST_FMT = 'MUCHGC{battery_max_grid_charging_current:03.0f}'


class MUCHGC_Parallel(MUCHGC):
    REQUEST_FMT = 'MUCHGC{_parallel_no:d}{battery_max_grid_charging_current:03.0f}'
    REQUEST_DEFAULT_VALUES = {
        '_parallel_no': 0
    }




from typing import Any, Dict, Union


Property = Union[str, int, float, bool]
Properties = Dict[str, Property]
PropertyDefinition = Dict[str, Dict[str, Any]]
PropertyDefinitions = Dict[str, PropertyDefinition]




import ctypes
import re

from typing import Optional, Tuple

from ..exceptions import MPPSolarException
from ..typing import Properties, PropertyDefinitions


SUFFIXES = ('__choices',)


class CommandException(MPPSolarException):
    pass


class ResponseError(CommandException):
    pass


class Command:
    REQUEST_FMT = ''
    RESPONSE_FMT = ''
    VIRTUAL_PROPERTIES = {}
    UNITS = {}
    DISPLAY_NAMES = {}
    CHOICES = {}
    REQUEST_DEFAULT_VALUES = {}

    _TYPE_MAP = {
        'int': int,
        'float': float,
        'bool': bool,
        'str': str
    }

    _response_regex: Optional[Tuple[Optional[re.Pattern], re.Pattern]] = None
    _request_property_definitions: Optional[PropertyDefinitions] = None
    _response_property_definitions: Optional[PropertyDefinitions] = None

    def __init__(self, **params) -> None:
        self._params = params

    @classmethod
    def get_name(cls) -> str:
        return cls.__name__.split('_')[0]

    def prepare_request(self) -> bytes:
        message = self.REQUEST_FMT.format(**self._params)
        return message.encode() + self.compute_crc(message) + b'\r'

    def parse_response(self, response: bytes) -> Properties:
        if not response.startswith(b'('):
            raise ResponseError(f'Unexpected response start: {response}')

        if not response.endswith(b'\r'):
            raise ResponseError(f'Unexpected response end: {response}')

        response = response[:-1]  # Get rid of terminal '\r'

        crc = response[-2:]
        response = response[:-2].decode()
        if self.compute_crc(response) != crc:
            raise ResponseError(f'Wrong CRC: {response} {repr(crc)[2:-1]}')

        response = response[1:]  # Get rid of start byte '('

        split_pat, values_pat = self.get_response_regex()
        if split_pat:
            parts = split_pat.split(response)
        else:
            parts = [response]

        parsed_dict = {}
        for part in parts:
            match = values_pat.match(part)
            if not match:
                raise ResponseError(f'Unexpected response format: {response}')

            for name, value in match.groupdict().items():
                type_, name = name.split('_', 1)
                parsed_value = self._TYPE_MAP[type_](value)
                if split_pat:
                    parsed_dict.setdefault(name, []).append(parsed_value)
                else:
                    parsed_dict[name] = parsed_value

        # Add virtual properties
        for name, details in self.VIRTUAL_PROPERTIES.items():
            if name in parsed_dict:
                continue  # Property already present

            parsed_dict[name] = details['value'](parsed_dict)

        return parsed_dict

    @classmethod
    def get_request_property_definitions(cls) -> PropertyDefinitions:
        if cls._request_property_definitions is None:
            matches = re.findall(r'{([^:]+):([.0-9dfbs]+)}', cls.REQUEST_FMT)
            cls._request_property_definitions = {
                name: {
                    'format': format_
                } for name, format_ in matches if not name.startswith('_')
            }

        return cls._request_property_definitions

    @classmethod
    def get_response_property_definitions(cls) -> PropertyDefinitions:
        if cls._response_property_definitions is None:
            is_list = False
            response_fmt = cls.RESPONSE_FMT
            if response_fmt.endswith('...'):
                is_list = True
                response_fmt = response_fmt[:-3]

            matches = re.findall(r'{([^:]+):([dfbs])}', response_fmt)
            matches += [(name, details['type']) for name, details in cls.VIRTUAL_PROPERTIES.items()]
            type_mapping = {
                'd': 'int',
                'f': 'float',
                'b': 'bool',
                's': 'str'
            }
            cls._response_property_definitions = {
                re.sub('|'.join(SUFFIXES), '', name): {
                    'type': type_mapping.get(type_, type_),
                    'is_list': is_list,
                    'is_choices': name.endswith('__choices'),
                    'unit': cls.UNITS.get(name),
                    'display_name': cls.DISPLAY_NAMES.get(name),
                    'choices': [
                        {'value': choice[0], 'display_name': choice[1]}
                        for choice in cls.CHOICES[name]
                    ] if name in cls.CHOICES else None
                } for name, type_ in matches if not name.startswith('_')
            }

        return cls._response_property_definitions

    @classmethod
    def has_response_properties(cls) -> bool:
        for details in cls.get_response_property_definitions().values():
            if not details['is_choices']:
                return True

        return False

    @classmethod
    def get_response_regex(cls) -> Tuple[Optional[re.Pattern], re.Pattern]:
        if cls._response_regex is None:
            pat = cls.RESPONSE_FMT
            is_list = False
            if pat.endswith('...'):
                pat = pat[:-3]
                is_list = True
            pat = re.sub('|'.join(SUFFIXES), '', pat)
            pat = re.sub(r'\s+', '\\\\s+', pat)
            pat = re.sub(r'{([^:]+):d}', '(?P<int_\\1>-?[0-9]+)', pat)
            pat = re.sub(r'{([^:]+):f}', '(?P<float_\\1>-?[0-9.]+)', pat)
            pat = re.sub(r'{([^:]+):b}', '(?P<bool_\\1>[01])', pat)
            pat = re.sub(r'{([^:]+):s}', '(?P<str_\\1>[^\\\\s]+)', pat)
            cls._response_regex = (re.compile(r'\s+') if is_list else None, re.compile(pat))

        return cls._response_regex

    @staticmethod
    def compute_crc(message: str) -> bytes:
        crc = 0
        crc_ta = [
            0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
            0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef
        ]

        for c in message:
            c = ord(c)

            t_da = ctypes.c_uint8(crc >> 8)
            da = t_da.value >> 4
            crc <<= 4
            index = da ^ (c >> 4)
            crc ^= crc_ta[index]
            t_da = ctypes.c_uint8(crc >> 8)
            da = t_da.value >> 4
            crc <<= 4
            index = da ^ (c & 0x0f)
            crc ^= crc_ta[index]

        crc_low = ctypes.c_uint8(crc).value
        crc_high = ctypes.c_uint8(crc >> 8).value

        if crc_low in (0x28, 0x0d, 0x0a):
            crc_low += 1

        if crc_high in (0x28, 0x0d, 0x0a):
            crc_high += 1

        return bytes((crc_high, crc_low))




import abc
import logging
import struct

from typing import Any, Dict, List

from qtoggleserver.lib import ble

from . import constants
from . import ports as mppsolarports
from .inverter import MPPSolarInverter
from .typing import Property


logger = logging.getLogger(__name__)


class BluetoothPort(ble.BLEPort, metaclass=abc.ABCMeta):
    READ_INTERVAL_MAX = 86400
    READ_INTERVAL_STEP = 1
    READ_INTERVAL_MULTIPLIER = 1


class BooleanPort(mppsolarports.BooleanPort, BluetoothPort):
    pass


class NumberPort(mppsolarports.NumberPort, BluetoothPort):
    pass


class StringPort(mppsolarports.StringPort, BluetoothPort):
    pass


class BluetoothMPPSolarInverter(MPPSolarInverter, ble.BLEPeripheral):
    STATUS1_HANDLE = 0x001D
    STATUS2_HANDLE = 0x0020
    PV1_STATUS_HANDLE = 0x0042

    logger = logger

    async def read_properties(self) -> None:
        data = await self.read(self.STATUS1_HANDLE)
        self.parse_status1_data(data)

        data = await self.read(self.STATUS2_HANDLE)
        self.parse_status2_data(data)

        data = await self.read(self.PV1_STATUS_HANDLE)
        self.parse_pv1_status_data(data)

    async def set_property(self, name: str, value: Property) -> None:
        pass  # TODO: implement me

    def parse_status1_data(self, data: bytes) -> None:
        (
            grid_voltage, grid_frequency, ac_output_voltage, ac_output_frequency,
            ac_output_apparent_power, ac_output_active_power, ac_output_load,
            bus_voltage, battery_voltage, battery_charging_current
        ) = struct.unpack('<HHHHHHHHHH', data)

        self._properties['grid_voltage'] = grid_voltage / 10
        self._properties['grid_frequency'] = grid_frequency / 10
        self._properties['ac_output_voltage'] = ac_output_voltage / 10
        self._properties['ac_output_frequency'] = ac_output_frequency / 10
        self._properties['ac_output_apparent_power'] = ac_output_apparent_power
        self._properties['ac_output_active_power'] = ac_output_active_power
        self._properties['ac_output_load'] = ac_output_load
        self._properties['bus_voltage'] = bus_voltage
        self._properties['battery_voltage'] = battery_voltage / 100
        self._properties['battery_charging_current'] = battery_charging_current

    def parse_status2_data(self, data: bytes) -> None:
        battery_soc, heat_sink_temp, _, _, _, _, mode, _, _, _ = struct.unpack('<HHHHHHHHHH', data)

        self._properties['battery_state_of_charge'] = battery_soc
        self._properties['heat_sink_temperature'] = heat_sink_temp
        self._properties['mode'] = chr(mode)

    def parse_pv1_status_data(self, data: bytes) -> None:
        _, _, _, _, _, pv_current, pv_voltage, pv_power, _, _ = struct.unpack('<HHHHHHHHHH', data)

        self._properties['pv_current'] = pv_current / 10
        self._properties['pv_voltage'] = pv_voltage / 10
        self._properties['pv_power'] = pv_power

    async def make_port_args(self) -> List[Dict[str, Any]]:
        return [
            {
                'driver': NumberPort,
                'property_name': 'grid_voltage',
                'display_name': 'Grid Voltage',
                'unit': 'V'
            },
            {
                'driver': NumberPort,
                'property_name': 'grid_frequency',
                'display_name': 'Grid Frequency',
                'unit': 'Hz'
            },
            {
                'driver': NumberPort,
                'property_name': 'ac_output_voltage',
                'display_name': 'AC Output Voltage',
                'unit': 'V'
            },
            {
                'driver': NumberPort,
                'property_name': 'ac_output_frequency',
                'display_name': 'AC Output Frequency',
                'unit': 'Hz'
            },
            {
                'driver': NumberPort,
                'property_name': 'ac_output_apparent_power',
                'display_name': 'AC Output Apparent Power',
                'unit': 'VA'
            },
            {
                'driver': NumberPort,
                'property_name': 'ac_output_active_power',
                'display_name': 'AC Output Active Power',
                'unit': 'W'
            },
            {
                'driver': NumberPort,
                'property_name': 'ac_output_load',
                'display_name': 'AC Output Load',
                'unit': '%'
            },
            {
                'driver': NumberPort,
                'property_name': 'battery_voltage',
                'display_name': 'Battery Voltage',
                'unit': 'V'
            },
            {
                'driver': NumberPort,
                'property_name': 'battery_charging_current',
                'display_name': 'Battery Charging Current',
                'unit': 'A'
            },
            {
                'driver': NumberPort,
                'property_name': 'battery_state_of_charge',
                'display_name': 'Battery State Of Charge',
                'unit': '%'
            },
            {
                'driver': NumberPort,
                'property_name': 'heat_sink_temperature',
                'display_name': 'Heat Sink Temperature',
                'unit': 'C'
            },
            {
                'driver': StringPort,
                'property_name': 'mode',
                'display_name': 'Inverter Mode',
                'choices': [{'value': c[0], 'display_name': c[1]} for c in constants.MODE_CHOICES]
            },
            {
                'driver': NumberPort,
                'property_name': 'pv_current',
                'display_name': 'PV Current',
                'unit': 'A'
            },
            {
                'driver': NumberPort,
                'property_name': 'pv_voltage',
                'display_name': 'PV Voltage',
                'unit': 'V'
            },
            {
                'driver': NumberPort,
                'property_name': 'pv_power',
                'display_name': 'PV Power',
                'unit': 'W'
            }
        ]




import asyncio
import contextlib
import logging
import math
import re

from typing import Any, Dict, List, Optional, Set, Type

from qtoggleserver.utils import json as json_utils

from . import commands, constants
from .io import BaseIO, HIDRawIO, SerialIO
from .inverter import MPPSolarInverter
from .ports import BooleanPort, NumberPort, StringPort
from .typing import Properties, Property


logger = logging.getLogger(__name__)


class SerialMPPSolarInverter(MPPSolarInverter):
    DEFAULT_BAUD = 2400
    CMD_WAIT = 0.5  # seconds
    BATTERY_FORCE_WAIT = 5  # seconds

    # Filter out properties that we don't really want exposed
    BLACKLISTED_PROPERTIES = {
        'has_sbu_priority',
        'is_configuration_status_changed',
        'is_scc_firmware_updated',
        'is_battery_voltage_too_steady_while_charging',
        'battery_voltage_offset_fans',
        'eeprom_version',
        'is_dustproof_installed'
    }

    logger = logger

    def __init__(
        self,
        *,
        serial_port: str,
        serial_baud: int = DEFAULT_BAUD,
        blacklist_properties: Optional[List[str]] = None,
        force_battery_discharge_min_soc: Optional[int] = None,
        force_battery_charge_grid_min_voltage: Optional[int] = None,
        **kwargs
    ) -> None:

        self._serial_port: str = serial_port
        self._serial_baud: int = serial_baud
        self._blacklist_properties: Set[str] = set(blacklist_properties or [])
        self._force_battery_discharge_min_soc: Optional[int] = force_battery_discharge_min_soc
        self._force_battery_charge_min_grid_voltage: Optional[int] = force_battery_charge_grid_min_voltage
        self._setter_command_classes_by_property: Dict[str, List[Type[commands.Command]]] = {}
        self._choices_by_property: Dict[str, List[Dict[str, Any]]] = {}
        self._command_lock: asyncio.Lock = asyncio.Lock()

        super().__init__(**kwargs)

        for cls in commands.get_command_classes(self._model):
            for name in cls.get_request_property_definitions():
                self._setter_command_classes_by_property.setdefault(name, []).append(cls)

    def make_io(self) -> BaseIO:
        if re.match(r'.*hidraw\d+', self._serial_port):
            return HIDRawIO(self._serial_port)
        else:
            return SerialIO(self._serial_port, self._serial_baud)

    async def run_command(self, io: BaseIO, cls: Type[commands.Command], **params) -> Properties:
        params = dict(params, **self.prepare_command_params(cls))
        if params:
            params_str = ', '.join(f'{k}={json_utils.dumps(v)}' for k, v in params.items())
            self.debug('running command %s(%s)', cls.get_name(), params_str)
        else:
            self.debug('running command %s', cls.get_name())

        cmd = cls(**params)
        # Don't run empty commands
        if not cmd.REQUEST_FMT:
            return {}

        request = cmd.prepare_request()
        async with self._command_lock:
            self.debug('sending "%s"', repr(request)[2:-1])
            io.write(request)
            response = await io.read(self.TIMEOUT)
            await asyncio.sleep(self.CMD_WAIT)
            self.debug('received "%s"', repr(response)[2:-1])

        parsed_response = cmd.parse_response(response)

        return parsed_response

    def prepare_command_params(self, cmd: Type[commands.Command]) -> Properties:
        return cmd.REQUEST_DEFAULT_VALUES

    async def read_properties(self) -> None:
        with contextlib.closing(self.make_io()) as io:
            for cls in commands.get_command_classes(self._model):
                if cls.has_response_properties():
                    try:
                        self._properties.update(await self.run_command(io, cls))
                    except Exception as e:
                        self.error('command %s failed: %s', cls.get_name(), e)
                        continue

    async def set_property(self, name: str, value: Property) -> None:
        cmd_classes = self._setter_command_classes_by_property[name]
        params = {name: value}
        old_value = self._properties.get(name)

        if old_value == value:
            return

        self.debug('setting property "%s" from %s to %s', name, json_utils.dumps(old_value), json_utils.dumps(value))

        with contextlib.closing(self.make_io()) as io:
            for cls in cmd_classes:
                await self.run_command(io, cls, **params)

            await self._handle_post_set_property(io, name, old_value, value)

    async def _handle_post_set_property(
        self,
        io: BaseIO, name: str,
        old_value: Optional[Property],
        new_value: Property
    ) -> None:
        # When changing output source priority, the user normally expects the inverter to switch to battery
        # charging/discharging mode right away, depending on the new setting. The inverter doesn't do this automatically
        # as it attempts to preserve the current charging/discharging battery phase, so we need to force it by
        # temporarily adjusting back-to-(dis)charging battery parameters.
        if name == 'output_source_priority':
            mode = self._properties.get('mode')
            soc = self._properties.get('battery_state_of_charge', 0)
            grid_voltage = self._properties.get('grid_voltage', 0)
            if (
                new_value == constants.OUTPUT_SOURCE_PRIORITY_SBG and
                mode == constants.MODE_GRID and
                self._force_battery_discharge_min_soc is not None and
                soc >= self._force_battery_discharge_min_soc
            ):
                self.info('forcing battery discharge')
                await self._force_battery_discharge(io)
            elif (
                old_value == constants.OUTPUT_SOURCE_PRIORITY_SBG and
                mode == constants.MODE_BATTERY and
                self._force_battery_charge_min_grid_voltage is not None and
                grid_voltage > self._force_battery_charge_min_grid_voltage
            ):
                self.info('forcing battery charge')
                await self._force_battery_charge(io)

    async def _force_battery_discharge(self, io: BaseIO) -> None:
        cmd_classes = self._setter_command_classes_by_property.get('battery_back_to_discharging_voltage', [])
        if not cmd_classes:
            self.warning('cannot force battery mode: command not available')
            return
        cls = cmd_classes[0]

        battery_voltage = self._properties.get('battery_voltage')
        if battery_voltage is None:
            self.warning('cannot force battery mode: battery voltage not available')
            return

        battery_back_to_discharging_voltage = self._properties.get('battery_back_to_discharging_voltage')
        if battery_back_to_discharging_voltage is None:
            self.warning('cannot force battery mode: battery back-to-discharging voltage not available')
            return

        temp_value = int(battery_voltage)
        if temp_value == battery_voltage:
            temp_value -= 1

        if temp_value >= battery_back_to_discharging_voltage:
            self.debug('adjusting battery back-to-discharging voltage not needed')
            return

        try:
            await self.run_command(io, cls, battery_back_to_discharging_voltage=temp_value)
            await asyncio.sleep(self.BATTERY_FORCE_WAIT)
        finally:
            await self.run_command(io, cls, battery_back_to_discharging_voltage=battery_back_to_discharging_voltage)

    async def _force_battery_charge(self, io: BaseIO) -> None:
        cmd_classes = self._setter_command_classes_by_property.get('battery_back_to_charging_voltage', [])
        if not cmd_classes:
            self.warning('cannot force battery mode: command not available')
            return
        cls = cmd_classes[0]

        battery_voltage = self._properties.get('battery_voltage')
        if battery_voltage is None:
            self.warning('cannot force battery mode: battery voltage not available')
            return

        battery_back_to_charging_voltage = self._properties.get('battery_back_to_charging_voltage')
        if battery_back_to_charging_voltage is None:
            self.warning('cannot force battery mode: battery back-to-charging voltage not available')
            return

        temp_value = int(math.ceil(battery_voltage))
        if temp_value == battery_voltage:
            temp_value += 1

        if temp_value <= battery_back_to_charging_voltage:
            self.debug('adjusting battery back-to-charging voltage not needed')
            return

        try:
            await self.run_command(io, cls, battery_back_to_charging_voltage=temp_value)
            await asyncio.sleep(self.BATTERY_FORCE_WAIT)
        finally:
            await self.run_command(io, cls, battery_back_to_charging_voltage=battery_back_to_charging_voltage)

    async def make_port_args(self) -> List[Dict[str, Any]]:
        # All available command classes for this inverter model
        cmd_classes = commands.get_command_classes(self._model)

        # Fetch property choices
        with contextlib.closing(self.make_io()) as io:
            for cls in cmd_classes:
                response_property_definitions = cls.get_response_property_definitions()
                for name, details in response_property_definitions.items():
                    if not details['is_choices']:
                        continue

                    response = await self.run_command(io, cls)
                    self._choices_by_property[name] = [
                        {
                            'value': c,
                            'label': str(c)
                        }
                        for c in response[name]
                    ]

        # Create port args
        blacklisted_properties = self.BLACKLISTED_PROPERTIES | self._blacklist_properties
        port_args_list = []
        for cls in cmd_classes:
            response_property_definitions = cls.get_response_property_definitions()
            for name, details in response_property_definitions.items():
                if (name in blacklisted_properties) or details['is_choices']:
                    continue

                type_ = details['type']
                port_args = {
                    'property_name': name,
                    'display_name': details['display_name'],
                    'writable': name in self._setter_command_classes_by_property
                }
                if type_ == 'bool':
                    port_args['driver'] = BooleanPort
                elif type_ in ('int', 'float'):
                    port_args['driver'] = NumberPort
                elif type_ == 'str':
                    port_args['driver'] = StringPort
                else:
                    continue

                if type_ in ('int', 'str'):
                    port_args.update({
                        'unit': details['unit'],
                        'choices': details['choices'] or self._choices_by_property.get(name),
                    })

                port_args_list.append(port_args)

        return port_args_list




import asyncio
import abc
import os
import serial


class BaseIO(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def read_available(self) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    def write(self, data: bytes) -> None:
        raise NotImplementedError

    async def read(self, timeout: int) -> bytes:
        data = b''
        for _ in range(timeout * 10):
            data += self.read_available()
            if data.endswith(b'\r'):
                break

            await asyncio.sleep(0.1)

        return data

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError


class HIDRawIO(BaseIO):
    def __init__(self, device_path: str) -> None:
        self._fd: int = os.open(device_path, flags=os.O_RDWR | os.O_NONBLOCK)

    def read_available(self) -> bytes:
        return os.read(self._fd, 1024)

    def write(self, data: bytes) -> None:
        os.write(self._fd, data)

    def close(self) -> None:
        os.close(self._fd)


class SerialIO(BaseIO):
    def __init__(self, serial_port: str, serial_baud: int) -> None:
        self._serial: serial.Serial = serial.Serial(serial_port, serial_baud)

    def read_available(self) -> bytes:
        in_waiting = self._serial.in_waiting
        if in_waiting:
            return self._serial.read(in_waiting)

        return b''

    def write(self, data: bytes) -> None:
        self._serial.write(data)

    def close(self) -> None:
        self._serial.close()




from .base import Command

from qtoggleserver.mppsolar import constants


class QPIRI(Command):
    REQUEST_FMT = 'QPIRI'

    UNITS = {
        'battery_back_to_charging_voltage': 'V',
        'battery_cut_off_voltage': 'V',
        'battery_bulk_charging_voltage': 'V',
        'battery_float_charging_voltage': 'V',
        'battery_max_grid_charging_current': 'A',
        'battery_max_charging_current': 'A',
        'battery_back_to_discharging_voltage': 'V',
        'battery_max_discharging_current': 'A',
    }

    DISPLAY_NAMES = {
        'battery_back_to_charging_voltage': 'Battery Back-to-charging Voltage',
        'battery_cut_off_voltage': 'Battery Cut-off Voltage',
        'battery_bulk_charging_voltage': 'Battery Bulk Charging Voltage',
        'battery_float_charging_voltage': 'Battery Float Charging Voltage',
        'battery_type': 'Battery Type',
        'battery_max_grid_charging_current': 'Battery Max Grid Charging Current',
        'battery_max_charging_current': 'Battery Max Charging Current',
        'output_source_priority': 'Output Source Priority',
        'charging_source_priority': 'Charging Source Priority',
        'battery_back_to_discharging_voltage': 'Battery Back-to-discharging Voltage',
        'battery_max_discharging_current': 'Battery Max Discharging Current',
    }

    CHOICES = {
        'battery_type': constants.BATTERY_TYPE_CHOICES,
        'output_source_priority': constants.OUTPUT_SOURCE_PRIORITY_CHOICES,
        'charging_source_priority': constants.CHARGING_SOURCE_PRIORITY_CHOICES,
        'battery_max_discharging_current': constants.BATTERY_MAX_DISCHARGING_CURRENT_CHOICES,
    }


class QPIRI_GK(QPIRI):
    RESPONSE_FMT = (
        '{_grid_rating_voltage:f} '
        '{_grid_rating_current:f} '
        '{_ac_output_rating_voltage:f} '
        '{_ac_output_rating_frequency:f} '
        '{_ac_output_rating_current:f} '
        '{_ac_output_rating_apparent_power:f} '
        '{_ac_output_rating_active_power:f} '
        '{_battery_rating_voltage:f} '
        '{battery_back_to_charging_voltage:f} '
        '{battery_cut_off_voltage:f} '
        '{battery_bulk_charging_voltage:f} '
        '{battery_float_charging_voltage:f} '
        '{battery_type:d} '
        '{battery_max_grid_charging_current:d} '
        '{battery_max_charging_current:d} '
        '{_input_voltage_range:d} '
        '{output_source_priority:d} '
        '{charging_source_priority:d} '
        '{_parallel_max:d} '
        '{_type:d} '
        '{_topology:d} '
        '{_output_mode:d} '
        '{battery_back_to_discharging_voltage:f} '
        '{_parallel_pv_ok:d} '
        '{_parallel_pv_power_balance:d}'
    )


class QPIRI_MK(QPIRI):
    RESPONSE_FMT = (
        '{_grid_rating_voltage:f} '
        '{_grid_rating_current:f} '
        '{_ac_output_rating_voltage:f} '
        '{_ac_output_rating_frequency:f} '
        '{_ac_output_rating_current:f} '
        '{_ac_output_rating_apparent_power:f} '
        '{_ac_output_rating_active_power:f} '
        '{_battery_rating_voltage:f} '
        '{battery_back_to_charging_voltage:f} '
        '{battery_cut_off_voltage:f} '
        '{battery_bulk_charging_voltage:f} '
        '{battery_float_charging_voltage:f} '
        '{battery_type:d} '
        '{battery_max_grid_charging_current:d} '
        '{battery_max_charging_current:d} '
        '{_input_voltage_range:d} '
        '{output_source_priority:d} '
        '{charging_source_priority:d} '
        '{_parallel_max:d} '
        '{_type:d} '
        '{_topology:d} '
        '{_output_mode:d} '
        '{battery_back_to_discharging_voltage:f} '
        '{_parallel_pv_ok:d} '
        '{_parallel_pv_power_balance:d} '
        '{_battery_max_cv_charging_time:d} '
        '{_operation_logic:d}'
    )


class QPIRI_MAX(QPIRI):
    RESPONSE_FMT = (
        '{_grid_rating_voltage:f} '
        '{_grid_rating_current:f} '
        '{_ac_output_rating_voltage:f} '
        '{_ac_output_rating_frequency:f} '
        '{_ac_output_rating_current:f} '
        '{_ac_output_rating_apparent_power:f} '
        '{_ac_output_rating_active_power:f} '
        '{_battery_rating_voltage:f} '
        '{battery_back_to_charging_voltage:f} '
        '{battery_cut_off_voltage:f} '
        '{battery_bulk_charging_voltage:f} '
        '{battery_float_charging_voltage:f} '
        '{battery_type:d} '
        '{battery_max_grid_charging_current:d} '
        '{battery_max_charging_current:d} '
        '{_input_voltage_range:d} '
        '{output_source_priority:d} '
        '{charging_source_priority:d} '
        '{_parallel_max:d} '
        '{_type:d} '
        '{_topology:d} '
        '{_output_mode:d} '
        '{battery_back_to_discharging_voltage:f} '
        '{_parallel_pv_ok:d} '
        '{_parallel_pv_power_balance:d} '
        '{_battery_max_cv_charging_time:d} '
        '{_operation_logic:d} '
        '{battery_max_discharging_current:d}'
    )




from setuptools import setup, find_namespace_packages


setup(
    name='qtoggleserver-mppsolar',
    version='1.4.5',
    description='MPP Solar inverters support for qToggleServer',
    author='Calin Crisan',
    author_email='ccrisan@gmail.com',
    license='Apache 2.0',

    packages=find_namespace_packages()
)




from .base import Command

from qtoggleserver.mppsolar import constants


class QMOD(Command):
    REQUEST_FMT = 'QMOD'
    RESPONSE_FMT = '{mode:s}'

    DISPLAY_NAMES = {
        'mode': 'Inverter Mode'
    }

    CHOICES = {
        'mode': constants.MODE_CHOICES
    }




MODE_POWER_ON = 'P'
MODE_STANDBY = 'S'
MODE_GRID = 'L'
MODE_BATTERY = 'B'
MODE_FAULT = 'F'
MODE_POWER_SAVING = 'H'
MODE_SHUTDOWN = 'D'
MODE_CHARGING = 'C'
MODE_BYPASS = 'Y'
MODE_ECO = 'E'

MODE_CHOICES = [
    (MODE_POWER_ON, 'Power On'),
    (MODE_STANDBY, 'Standby'),
    (MODE_GRID, 'Grid'),
    (MODE_BATTERY, 'Battery'),
    (MODE_FAULT, 'Fault'),
    (MODE_POWER_SAVING, 'Power Saving'),
    (MODE_SHUTDOWN, 'Shutdown'),
    (MODE_CHARGING, 'Charging'),
    (MODE_BYPASS, 'Bypass'),
    (MODE_ECO, 'Eco')
]

BATTERY_TYPE_AGM = 0
BATTERY_TYPE_FLOODED = 1
BATTERY_TYPE_USER = 2
BATTERY_TYPE_PYLONTECH = 3

BATTERY_TYPE_CHOICES = [
    (BATTERY_TYPE_AGM, 'AGM'),
    (BATTERY_TYPE_FLOODED, 'Flooded'),
    (BATTERY_TYPE_USER, 'User'),
    (BATTERY_TYPE_PYLONTECH, 'Pylontech'),
]

OUTPUT_SOURCE_PRIORITY_GSB = 0
OUTPUT_SOURCE_PRIORITY_SGB = 1
OUTPUT_SOURCE_PRIORITY_SBG = 2

OUTPUT_SOURCE_PRIORITY_CHOICES = [
    (OUTPUT_SOURCE_PRIORITY_GSB, 'Grid/Solar/Battery'),
    (OUTPUT_SOURCE_PRIORITY_SGB, 'Solar/Grid/Battery'),
    (OUTPUT_SOURCE_PRIORITY_SBG, 'Solar/Battery/Grid'),
]

CHARGING_SOURCE_PRIORITY_STG = 1
CHARGING_SOURCE_PRIORITY_SAG = 2
CHARGING_SOURCE_PRIORITY_OS = 3

CHARGING_SOURCE_PRIORITY_CHOICES = [
    (0, 'Only Grid'),
    (CHARGING_SOURCE_PRIORITY_STG, 'Solar Then Grid'),
    (CHARGING_SOURCE_PRIORITY_SAG, 'Solar And Grid'),
    (CHARGING_SOURCE_PRIORITY_OS, 'Only Solar'),
]

BATTERY_MAX_DISCHARGING_CURRENT_CHOICES = [
    (0, 'Disabled'),
    (30, '30'),
    (40, '40'),
    (50, '50'),
    (60, '60'),
    (70, '70'),
    (80, '80'),
    (90, '90'),
    (100, '100'),
    (110, '110'),
    (120, '120'),
    (130, '130'),
    (140, '140'),
    (150, '150'),
]




from .base import Command


class MNCHGC(Command):
    pass


class MNCHGC_Single(MNCHGC):
    REQUEST_FMT = 'MNCHGC{battery_max_charging_current:03.0f}'


class MNCHGC_Parallel(MNCHGC):
    REQUEST_FMT = 'MNCHGC{_parallel_no:d}{battery_max_charging_current:03.0f}'
    REQUEST_DEFAULT_VALUES = {
        '_parallel_no': 0
    }
